#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <sys/mman.h>
#include <poll.h>
#include <errno.h>
#include <dirent.h>

#include "config.h"

#include <archive.h>
#include <archive_entry.h>

#ifdef HAVE_UDEV
#include <libudev.h>
#endif

#include "partition.h"
#include "version.h"
#include "odin3-proto.h"

#define USB_VENDOR_SAMSUNG "04e8"
#define USB_DEVICE_Z100 "6601"
#define USB_DEVICE_Z100_DEVGURU "6860"
#define USB_DEVICE_OLD_DOWNLOADER "1204"

#define KB			(1024)
#define MB			(1024*KB)
#define DEFAULT_PKT_SIZE	(128*KB)	/* sizeof the download unit (no
						   more, no less) */
#define DEFAULT_TIMEOUT		(1000)		/* 1 second */
#define DEFAULT_PORT		"/dev/ttyACM0"
#define PKT_DUMP_SIZE		(500)		/* number of bytes we get in one
						   go when receiving a PIT file
						   */
#define MAX_PIT_SIZE		(0x10000)	/* maximum expected sizeof PIT */

/* data abstraction */
struct data_src {
	size_t (*get_data_length)(struct data_src *src);
	size_t (*get_data_block)(struct data_src *src, void *data, size_t len);
};

int opt_verbose = 0;
int opt_test = 0;
size_t trans_unit_size = DEFAULT_PKT_SIZE;

static void dump(const char *msg, void *bytes, int len)
{
	int i;

	if (!opt_verbose)
		return;

	fprintf(stderr, "%s :", msg);
	for (i=0; i<len; i++)
	{
		if (i && (0==i%16))
			fprintf(stderr, "      ");
		fprintf(stderr, "%02x%c", ((unsigned char*)bytes)[i], 
				((len == i+1) || (i+1)%16==0) ? '\n': ' ');
	}
}

int serial_read(int fd, void *addr, size_t len, int timeout)
{
	unsigned char *p = addr;
	int n = 0, r;

	while (n < len)
	{
		struct pollfd fds;
		fds.fd = fd;
		fds.events = POLLIN;
		fds.revents = 0;
		r = poll(&fds, 1, timeout);
		if (r < 0)
		{
			fprintf(stderr, "line %d: poll failed\n", __LINE__);
			return -1;
		}

		if (r == 0)
		{
			if (opt_verbose)
				fprintf(stderr, "line %d: timeout after reading %d bytes\n", __LINE__, n);
			return n;
		}

		if (!(fds.revents & POLLIN))
		{
			fprintf(stderr, "line %d: poll returned non-read event %08x\n", __LINE__, fds.revents);
			continue;
		}

		r = read(fd, &p[n], len - n);
		if (r < 0)
		{
			fprintf(stderr, "line %d: read failed\n", __LINE__);
			return -1;
		}

		n += r;
	}

	return n;
}


/* for every 128k we send, we expect a response */
int send_chunks(int fd, struct data_src *senddata, size_t size)
{
	unsigned char *chunk;
	struct response_packet resp;
	size_t n = 0;
	int chunksize, chunk_number = 0;
	struct timeval stv, etv;
	double elaps;
	int r;

	chunk = malloc(trans_unit_size);
	if (chunk == NULL)
	{
		fprintf(stderr, "Unable to allocate memory.\n");
		return -1;
	}

	if (opt_verbose)
		fprintf(stderr, "Downloading: %zd bytes\n", size);

	gettimeofday(&stv, NULL);

	while (n < size)
	{
		if (size - n >= trans_unit_size)
			chunksize = trans_unit_size;
		else
			chunksize = size - n;

		chunksize = senddata->get_data_block(senddata, chunk, chunksize);
		if (chunksize <= 0)
		{
			fprintf(stderr, "\nline %d: Broken tar archive - check download!\n", __LINE__);
			exit(1);
		}

		if (opt_verbose)
			fprintf(stderr, "read %d bytes\n", chunksize);

		assert(chunksize > 0 && chunksize <= trans_unit_size);
		memset(&chunk[chunksize], 0, trans_unit_size - chunksize);

		if (opt_verbose)
			fprintf(stderr, "sending %zd/%zd\n", n, size);

		r = write(fd, chunk, chunksize);
		if (r != chunksize)
		{
			fprintf(stderr, "line %d: failed to send data\n", __LINE__);
			free(chunk);
			return -1;
		}

		memset(&resp, 0, sizeof resp);
		r = serial_read(fd, &resp, sizeof resp, DEFAULT_TIMEOUT);

		/*
		 * Old bootloaders want a full chunk every time
		 * If we timeout waiting for a response on the last partial chunk
		 *  assume the bootloader wants a full chunk and send the rest
		 */
		if (r == 0 && chunksize != trans_unit_size)
		{
			r = write(fd, chunk, trans_unit_size - chunksize);
			if (r != (trans_unit_size - chunksize))
			{
				fprintf(stderr, "line %d: failed to send data\n", __LINE__);
				free(chunk);
				return -1;
			}
			r = serial_read(fd, &resp, sizeof resp, DEFAULT_TIMEOUT);
		}

		if (r != sizeof resp)
		{
			fprintf(stderr, "line %d: read failed %d\n", __LINE__, r);
			free(chunk);
			return -1;
		}

		if (resp.ack != chunk_number)
		{
			printf("\nsending %zd/%zd : sequence wrong %08x, expected %08x\n",
				n, size, resp.ack, chunk_number);
			//return -1;
		}

		/* first chunk is 0 */
		n += chunksize;

		if (opt_verbose)
		{
			printf("response %08x %08x\n", resp.id, resp.ack);

		}
		else
		{
			int now = n/KB;
			int total = size/KB;
			char progress [4] = { '-', '\\', '|', '/' };
			char c = progress[(now/30)%4];
			if (n >= size)
			{
				gettimeofday(&etv, NULL);
				if (etv.tv_usec < stv.tv_usec)
				{
					elaps = etv.tv_sec - stv.tv_sec - 1;
					elaps += (double)(1*1000*1000 - stv.tv_usec + etv.tv_usec)/(1000*1000);
				}
				else
				{
					elaps = etv.tv_sec - stv.tv_sec;
					elaps += (double)(etv.tv_usec - stv.tv_usec)/(1000*1000);
				}
				fprintf(stderr, "\x1b[1A\x1b[16C%c sending %6dk/%6dk %3d%% block %-6d", 
						c, now, total, ((now*100)/total), chunk_number);
				fprintf(stderr, " [%.2f MB/s]\n", (double)(size/elaps)/(MB));
			}
			else
			{
				fprintf(stderr, "\x1b[1A\x1b[16C%c sending %6dk/%6dk %3d%% block %-6d\n", 
						c, now, total, ((now*100)/total), chunk_number);
			}
		}
		chunk_number++;
	}

	free(chunk);

	return 0;
}

/*
 * function that doesn't wait for response data, used by standard requests and
 * REQ_PIT_DATA_START, which doesn't want a response
 */
int write_request_to_wire(int fd, request_type req_id, int req_sub_id, int *data, int count)
{
	struct request_packet req;
	int r, i;

	assert(count <= sizeof(req.int_data)/sizeof(req.int_data[0]));
	assert(count >= 0);

	memset(&req, 0, sizeof(req));

	req.id = req_id;
	req.sub_id = req_sub_id;

	if (data)
	{
		for (i=0; i<count; i++)
			req.int_data[i] = data[i];
	}

	dump("send", &req, 0x20);

	r = write(fd, &req, INIT_REQUEST_SIZE);
	if (r != INIT_REQUEST_SIZE)
	{
		fprintf(stderr, "line %d: failed to send data r = %d\n", __LINE__, r);
		return -1;
	}

	return 0;
}

/* prototype from documentation's Request() */
int send_request_timeout(int fd, request_type req_id, int req_sub_id, int *data, int count, int timeout)
{
	struct response_packet resp;
	int r;

	r = write_request_to_wire(fd, req_id, req_sub_id, data, count);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to write request r = %d\n", __LINE__, r);
		return -1;
	}

	r = serial_read(fd, &resp, sizeof resp, timeout);
	if (r >= 0)
		dump("recv", &resp, sizeof resp);
	if (r != sizeof resp)
	{
		fprintf(stderr, "line %d: failed to receive data r = %d\n", __LINE__, r);
		return -1;
	}

	return resp.ack;
}

int send_request(int fd, request_type req_id, int req_sub_id, int *data, int count)
{
	return send_request_timeout(fd, req_id, req_sub_id, data, count, DEFAULT_TIMEOUT);
}

int write_pit_data(int fd, void *pitaddr, int pitsize)
{
	struct response_packet resp;
	int r;

	r = send_request(fd, RQT_PIT, RQT_PIT_DATA_SET, NULL, 0);
	if (r)
	{
		fprintf(stderr, "RQT_PIT_DATA_SET,        status = %08x\n", r);
		return -1;
	}

	r = send_request(fd, RQT_PIT, RQT_PIT_DATA_START, &pitsize, 1);
	if (r)
	{
		fprintf(stderr, "RQT_PIT_DATA_START,      status = %08x\n", r);
		return -1;
	}

	r = write(fd, pitaddr, pitsize);
	if (r != pitsize)
	{
		fprintf(stderr, "line %d: failed to send file\n", __LINE__);
		return -1;
	}
	r = serial_read(fd, &resp, sizeof resp, DEFAULT_TIMEOUT*10);
	if (r >= 0)
		dump("recv", &resp, sizeof resp);
	if (r != sizeof resp)
	{
		fprintf(stderr, "line %d: failed to receive data r = %d\n", __LINE__, r);
		return -1;
	}

	r = send_request(fd, RQT_PIT, RQT_PIT_DATA_COMPLETE, &pitsize, 1);
	printf("RQT_PIT_DATA_COMPLETE,   status = %08x\n", r);
	if (r < 0)
	{
		fprintf(stderr, "line %d: partitioning failed\n", __LINE__);
		return -1;
	}

	return 0;
}

int send_part_data(int fd, struct data_src *senddata, size_t size, int final, partition_info *part)
{
	int data[7];
	int r;
	struct timeval stv, etv;
	double elaps;

	/* announce we're going to send the next chunk */
	r = send_request(fd, RQT_XMIT, RQT_XMIT_DATA_DOWNLOAD, (int*)&size, 1);
	if (r)
	{
		fprintf(stderr, "RQT_XMIT_DATA_DOWNLOAD,  status = %08x\n", r);
		return -1;
	}

	r = send_request(fd, RQT_XMIT, RQT_XMIT_DATA_START, (int*)&size, 1);
	if (r)
	{
		fprintf(stderr, "RQT_XMIT_DATA_START,     status = %08x\n", r);
		return -1;
	}

	r = send_chunks(fd, senddata, size);
	if (r < 0)
	{
		fprintf(stderr, "line %d: sending failed\n", __LINE__);
		return -1;
	}

	fprintf(stderr, "\tWriting block... ");

	gettimeofday(&stv, NULL);

	data[0] = part->nBinType;
	if (part->nBinType == 0)
	{
		/* AP */
		data[1] = size;
		data[2] = part->nBinType;
		data[3] = part->nDevType;
		data[4] = part->nID;
		data[5] = final;
	}
	else
	{
		/* CP */
		data[1] = size;
		data[2] = 0;	// efs_clear	<- CSA
		data[3] = 0;	// boot_update	<- not used in SLP
		data[4] = final;
		data[5] = 0;
	}
	data[6] = 0;

	/* flashing takes some time... be patient here */
	r = send_request_timeout(fd, RQT_XMIT, RQT_XMIT_DATA_COMPLETE, data,
			sizeof(data)/sizeof(int) , 60000);
	if (r < 0)
		fprintf(stderr, "timeout waiting for flashing to complete\n");

	gettimeofday(&etv, NULL);
	if (etv.tv_usec < stv.tv_usec)
	{
		elaps = etv.tv_sec - stv.tv_sec - 1;
		elaps += (double)(1*1000*1000 - stv.tv_usec + etv.tv_usec)/(1000*1000);
	}
	else
	{
		elaps = etv.tv_sec - stv.tv_sec;
		elaps += (double)(etv.tv_usec - stv.tv_usec)/(1000*1000);
	}

	fprintf(stderr, "complete [%.2f MB/s]\n", (double)(size/elaps)/(MB));

	return 0;
}

int request_reboot(int fd)
{
	int r;
	r = send_request(fd, RQT_CLOSE, RQT_CLOSE_DATA_REBOOT, NULL, 0);
	if (r)
		fprintf(stderr, "RQT_CLOSE_DATA_REBOOT,   status = %08x\n", r);
	return r;
}

int odin_handshake(int fd)
{
	char buffer[4];
	int r;

	r = write(fd, "ODIN", 4);
	if (r != 4)
	{
		fprintf(stderr, "line %d: failed to write signature bytes\n", __LINE__);
		return -1;
	}

	r = serial_read(fd, buffer, 4, DEFAULT_TIMEOUT);
	if (r != 4)
	{
		fprintf(stderr, "line %d: failed to read signature bytes\n", __LINE__);
		return -1;
	}

	if (memcmp(buffer, "LOKE", 4))
	{
		fprintf(stderr, "line %d: signature byte mismatch\n", __LINE__);
		return -1;
	}

	return 0;
}

#ifndef HAVE_UDEV

int getfile(const char *devpath, char *buffer, size_t bufsize)
{
	int fd, r = 0;

	fd = open(devpath, O_RDONLY);
	if (fd >= 0)
	{
		r = read(fd, buffer, bufsize - 1);
		if (r < 0)
			r = 0;
		close(fd);
	}
	else if (opt_verbose)
		fprintf(stderr, "failed to open %s\n", devpath);
	if (r && buffer[r - 1] == '\n')
		r--;
	buffer[r] = 0;
	return r;
}

/*
 * Given USB path, return tty name
 * FIXME: probably should look up the /dev from device major:minor
 * There seems to be two styles of device name:
 *  1. /sys/bus/usb/devices/1-2/1-2\:2.0/tty/ttyACM0/dev
 *  2. /sys/bus/usb/devices/1-2/1-2\:2.0/tty:ttyACM0
 */
char *device_from_usb_tty_directory(const char *usbpath)
{
	static char ttypath[0x400];
	static char devpath[0x400];
	char *ret = NULL;
	DIR *d;

	strcpy(ttypath, usbpath);
	strcat(ttypath, "/tty");

	d = opendir(ttypath);
	if (d)
	{
		if (opt_verbose)
			fprintf(stderr, "listing %s:\n", ttypath);

		/* device name is under tty directory */
		while (!ret)
		{
			struct dirent *de;

			de = readdir(d);
			if (!de)
				break;

			if (opt_verbose)
				fprintf(stderr, "%s\n", de->d_name);

			if (de->d_name[0] == '.')
				continue;

			strcpy(devpath, "/dev/");
			strcat(devpath, de->d_name);

			ret = devpath;
		}
	}
	else
	{
		/* device name is in directory */
		d = opendir(usbpath);
		if (!d)
			return NULL;

		if (opt_verbose)
			fprintf(stderr, "listing %s:\n", usbpath);

		while (!ret)
		{
			struct dirent *de;

			de = readdir(d);
			if (!de)
				break;

			if (opt_verbose)
				fprintf(stderr, "%s\n", de->d_name);

			if (strncmp(de->d_name, "tty:", 4))
				continue;

			strcpy(devpath, "/dev/");
			strcat(devpath, &de->d_name[4]);

			ret = devpath;
		}
	}

	closedir(d);

	fprintf(stderr, "Device is %s\n\n", ret);

	return ret;
}

const char* find_usb_device(void)
{
	DIR *d;
	char *p;
	char buffer[11];
	const char *dirname = "/sys/bus/usb/devices";
	char usbpath[0x400];

	d = opendir(dirname);
	if (!d)
		return NULL;

	while (1)
	{
		struct dirent *de;

		de = readdir(d);
		if (!de)
			break;

		if (de->d_name[0] == '.')
			continue;

		strcpy(usbpath, dirname);
		strcat(usbpath, "/");
		strcat(usbpath, de->d_name);
		strcat(usbpath, "/");
		p = &usbpath[strlen(usbpath)];

		/* match the vendor ID */
		strcat(p, "idVendor");
		getfile(usbpath, buffer, sizeof buffer);
		if (opt_verbose)
			fprintf(stderr, "vendorId = >%s< %zd\n", buffer, strlen(buffer));
		if (strcmp(buffer, USB_VENDOR_SAMSUNG))
			continue;

		/* match the product ID */
		strcpy(p, "idProduct");
		getfile(usbpath, buffer, sizeof buffer);
		if (opt_verbose)
			fprintf(stderr, "product = >%s< %zd\n", buffer, strlen(buffer));

		if (!strcmp(buffer, USB_DEVICE_OLD_DOWNLOADER))
			fprintf(stderr, "Old bootloader detected!\n");

		if (strcmp(buffer, USB_DEVICE_Z100) &&
		    strcmp(buffer, USB_DEVICE_Z100_DEVGURU))
			continue;

		if (opt_verbose)
			fprintf(stderr, "found %s:%s!\n", USB_VENDOR_SAMSUNG, buffer);

		strcpy(p, de->d_name);
		strcat(p, ":2.0");

		return device_from_usb_tty_directory(usbpath);
	}

	closedir(d);

	return NULL;
}

#else

char *tty_from_udev_device(struct udev_device *dev)
{
	const char *model_id, *vendor_id;
	const char *devname;

	model_id = udev_device_get_property_value(dev, "ID_MODEL_ID");
	vendor_id = udev_device_get_property_value(dev, "ID_VENDOR_ID");

	if (strcmp(vendor_id, USB_VENDOR_SAMSUNG))
		return NULL;

	if (!strcmp(model_id, USB_DEVICE_OLD_DOWNLOADER))
	{
		fprintf(stderr, "Incompatible legacy downloader detected!\n");
		return NULL;
	}

	if (strcmp(model_id, USB_DEVICE_Z100) &&
	    strcmp(model_id, USB_DEVICE_Z100_DEVGURU))
		return NULL;


	if (opt_verbose)
		fprintf(stderr, "found %s:%s!\n", vendor_id, model_id);

	devname = udev_device_get_property_value(dev, "DEVNAME");
	if (!devname)
		return NULL;

	return strdup(devname);
}

const char* find_usb_device(void)
{
	struct udev *udev;
	struct udev_monitor *umon;
	struct udev_enumerate *uenum;
	struct udev_list_entry *ulist;
	struct udev_device *dev;
	char *ttypath = NULL;

	udev = udev_new();
	if (!udev)
	{
		fprintf(stderr, "failed to init udev\n");
		return NULL;
	}

	/*
	 * check if device is already plugged
	 */
	uenum = udev_enumerate_new(udev);
	if (!uenum)
	{
		fprintf(stderr, "failed to init udev enumerator\n");
		return NULL;
	}

	udev_enumerate_add_match_subsystem(uenum, "tty");
	udev_enumerate_add_match_property(uenum, "ID_VENDOR_ID", USB_VENDOR_SAMSUNG);

	udev_enumerate_scan_devices(uenum);

	ulist = udev_enumerate_get_list_entry(uenum);
	while (ulist && !ttypath)
	{
		const char *name = udev_list_entry_get_name(ulist);

		dev = udev_device_new_from_syspath(udev, name);
		if (dev)
			ttypath = tty_from_udev_device(dev);

		ulist = udev_list_entry_get_next(ulist);
	}
	udev_enumerate_unref(uenum);

	if (ttypath)
	{
		udev_unref(udev);
		return ttypath;
	}

	/*
	 * wait for a device add event
	 */
	umon = udev_monitor_new_from_netlink(udev, "udev");
	if (!umon)
	{
		fprintf(stderr, "failed to init udev monitor\n");
		return NULL;
	}

	udev_monitor_enable_receiving(umon);
	while (!ttypath)
	{
		const char *subsys;
		const char *action;

		dev = udev_monitor_receive_device(umon);
		if (!dev)
			continue;

		subsys = udev_device_get_subsystem(dev);
		action = udev_device_get_action(dev);
		if (!strcmp(subsys, "tty") && !strcmp(action, "add"))
		{
			ttypath = tty_from_udev_device(dev);
		}

		udev_device_unref(dev);
	}

	udev_monitor_unref(umon);
	udev_unref(udev);

	return ttypath;
}

#endif

/* TODO: this could be improved by looking for a USB device 04e8:6601 plug event, then opening it's ACM port */
int wait_and_open_port(const char *portname)
{
	int once = 0;
	int fd;
	int r;
	struct termios tios;
	const char *dev;

	if (opt_test)
		return open("/dev/null", O_RDWR);

	while (1)
	{
		if (!portname)
			dev = find_usb_device();
		else
			dev = portname;

		if (dev)
		{
			fd = open(dev, O_RDWR);
			if (fd >= 0)
				break;

			/* seem to get EIO with some dodgy 2.6.28 Linux kernels in Ubuntu... kernel bug? */
			if (errno == EIO)
			{
				fprintf(stderr, "open(%s) return EIO... kernel or phone's bootloader needs upgrade?\n", dev);
				return -1;
			}
		}

		if (!once)
		{
			printf("Port %s doesn't exist yet... reboot phone, type \"usb\" at u-boot prompt\n", portname ? portname : "");
			printf("or press <volume-down> + <camera button> + <power on> and cycle power\n\n");
			once = 1;
		}

		sleep(1);
	}

	r = tcgetattr(fd, &tios);
	if (r < 0)
	{
		fprintf(stderr, "line %d: tcgetattr failed\n", __LINE__);
		return -1;
	}

	/*
	 * Firmware BUG alert!!!
	 * Flow control (RTS/CTS) should be disabled, because
	 *  the firmware doesn't deal with it properly.
	 */
	cfmakeraw(&tios);
	r = tcsetattr(fd, TCSANOW, &tios);
	if (r < 0)
	{
		fprintf(stderr, "line %d: tcsetattr failed\n", __LINE__);
		return -1;
	}

	r = tcflush(fd, TCIOFLUSH);
	if (r < 0)
	{
		fprintf(stderr, "line %d: tcflush failed\n", __LINE__);
		return -1;
	}

	r = odin_handshake(fd);
	if (r < 0)
	{
		fprintf(stderr, "line %d: handshake failed\n", __LINE__);
		return -1;
	}

	return fd;
}

int write_data_to_file(const char *pitfile, void *pit, int len)
{
	int fd, r;

	fd = creat(pitfile, 0666);
	if (fd < 0)
	{
		fprintf(stderr, "line %d: failed to create %s: %d\n", __LINE__, pitfile, errno);
		return fd;
	}

	r = write(fd, pit, len);

	close(fd);

	return r;
}

int get_pit_data(int fd, unsigned char **data, int *data_length)
{
	int r, len, n, chunksize;
	unsigned char *pit = NULL;

	/* longer timeout to account for dumping data to the console */
	len = send_request_timeout(fd, RQT_PIT, RQT_PIT_DATA_GET, NULL, 0, 1000);
	if (len < 0)
	{
		fprintf(stderr, "line %d: get pit request failed %d\n", __LINE__, len);
		return -1;
	}

	if (opt_verbose)
		printf("pit is %d bytes\n", len);

	if (len > MAX_PIT_SIZE)
	{
		fprintf(stderr, "line %d: pit seems too big\n", __LINE__);
		return -1;
	}

	pit = malloc(len);
	if (pit == NULL)
	{
		fprintf(stderr, "Unable to allocate memory.\n");
		return -1;
	}

	n = 0;
	while (n*PKT_DUMP_SIZE < len)
	{
		r = write_request_to_wire(fd, RQT_PIT, RQT_PIT_DATA_START, &n, 1);
		if (r < 0)
		{
			fprintf(stderr, "line %d: get pit data start failed %d\n", __LINE__, r);
			return -1;
		}

		chunksize = (len - n*PKT_DUMP_SIZE);
		if (chunksize > PKT_DUMP_SIZE)
			chunksize = PKT_DUMP_SIZE;

		r = serial_read(fd, &pit[n*PKT_DUMP_SIZE], chunksize, 1000);
		dump("pit ", &pit[n*PKT_DUMP_SIZE], r);
		if (r != chunksize)
		{
			fprintf(stderr, "line %d: read returned %d/%d\n", __LINE__, r, chunksize);
			return -1;
		}
		n++;
	}

	r = send_request(fd, RQT_PIT, RQT_PIT_DATA_COMPLETE, NULL, 0);
	if (r < 0)
	{
		fprintf(stderr, "line %d: get pit data start failed %d\n", __LINE__, r);
		return -1;
	}

	*data = pit;
	*data_length = len;

	return 0;
}

int get_pit_file(const char *portname, const char *pitfile)
{
	unsigned char *pit = NULL;
	int pit_length = 0;
	int r;
	int fd;

	fd = wait_and_open_port(portname);
	if (fd < 0)
	{
		fprintf(stderr, "line %d: failed to open port %s\n", __LINE__, portname);
		return -1;
	}

	r = get_pit_data(fd, &pit, &pit_length);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to fetch pit data %d\n", __LINE__, r);
		return -1;
	}

	r = write_data_to_file(pitfile, pit, pit_length);
	if (r < 0)
	{
		fprintf(stderr, "line %d: save downloaded pit data failed %d\n", __LINE__, r);
		return -1;
	}

	free(pit);

	request_reboot(fd);

	close(fd);

	return 0;
}

size_t get_chunks_per_step(partition_info *part)
{
	switch (part->nDevType)
	{
        case PARTITION_DEV_TYPE_NAND:		/* FILE type include */
	case PARTITION_DEV_TYPE_ONENAND:
		return 100 * MB / trans_unit_size;	/* 104,857,600 bytes */
        case PARTITION_DEV_TYPE_MOVINAND:
		return 29440 * KB / trans_unit_size;	/* 30,146,560 bytes */
	default:
		fprintf(stderr, "unknown device type %d\n", part->nDevType);
		assert(0);
	}
}

/*
 * Write a data source (file) into a partition
 *
 * data_src abstraction provided so sending a single file
 * from a non-tar source is possible one day
 */
int flash_data(int fd, struct data_src *senddata, partition_info *part)
{
	size_t datasize, n, stepsize, numsteps;
	int r;

	datasize = senddata->get_data_length(senddata);

	stepsize = get_chunks_per_step(part) * trans_unit_size;

	numsteps = (datasize + stepsize - 1)/stepsize;
	if (opt_verbose)
		fprintf(stderr, "downloading in %zd steps\n", numsteps);

	for (n = 0; n < datasize; n += stepsize)
	{
		/* complete when the next step takes us to the size of the tarfile */
		int final = (n + stepsize) >= datasize;
		int len;

		len = datasize - n;
		if (len > stepsize)
			len = stepsize;

		fprintf(stderr, "step %zd/%zd\n", 1+n/stepsize, numsteps);

		r = send_part_data(fd, senddata, len, final, part);
		if (r < 0)
		{
			fprintf(stderr, "line %d: failed write tar file\n", __LINE__);
			return -1;
		}
	}

	fprintf(stderr, "\n");

	return 0;
}

int check_partition(unsigned char *pit, int pit_length)
{
	partition_header *hdr;

	if (pit_length < sizeof *hdr)
	{
		fprintf(stderr, "line %d: partition table too short\n", __LINE__);
		return -1;
	}

	hdr = (void*) pit;

	if (hdr->nCount <= 0)
	{
		fprintf(stderr, "line %d: no partitions\n", __LINE__);
		return -1;
	}

	if ((hdr->nCount * sizeof (partition_info) + sizeof *hdr) > pit_length)
	{
		fprintf(stderr, "line %d: too many partitions \n", __LINE__);
		return -1;
	}

	return 0;
}


partition_info* find_partition_by_filename(unsigned char *pit, int pit_length, const char *filename)
{
	partition_header *hdr;
	int i;

	if (0 > check_partition(pit, pit_length))
		return NULL;
	hdr = (void*) pit;

	for (i=0; i<hdr->nCount; i++)
	{
		partition_info *part = (void*) &pit[sizeof *hdr + i * sizeof (partition_info)];

		if (!strcmp(filename, part->szFileName))
			return part;
	}

	return NULL;
}

partition_info* find_partition_by_name(unsigned char *pit, int pit_length, const char *name)
{
	partition_header *hdr;
	int i;

	if (0 > check_partition(pit, pit_length))
		return NULL;
	hdr = (void*) pit;

	for (i=0; i<hdr->nCount; i++)
	{
		partition_info *part = (void*) &pit[sizeof *hdr + i * sizeof (partition_info)];

		if (!strcmp(name, part->szName))
			return part;
	}

	return NULL;
}

struct tar_data_src {
	struct data_src src;
	struct archive *ar;
	struct archive_entry *ae;
};

size_t te_get_data_length(struct data_src *src)
{
	struct tar_data_src *tardata = (struct tar_data_src *) src;

	return archive_entry_size(tardata->ae);
}

size_t te_get_data_block(struct data_src *src, void *data, size_t len)
{
	struct tar_data_src *tardata = (struct tar_data_src *) src;

	return archive_read_data(tardata->ar, data, len);
}

int tardata_open(struct tar_data_src *tardata, const char *tarfile)
{
	int r;

	/* open the tar archive */
	tardata->ar = archive_read_new();

	archive_read_support_format_tar(tardata->ar);
	archive_read_support_compression_gzip(tardata->ar);
	archive_read_support_compression_bzip2(tardata->ar);

	if (!strcmp(tarfile, "-"))
		r = archive_read_open_FILE(tardata->ar, stdin);
	else
		r = archive_read_open_filename(tardata->ar, tarfile, 512);

	if (r)
	{
		fprintf(stderr, "line %d: failed to open %s\n", __LINE__, tarfile);
		return r;
	}

	tardata->src.get_data_length = &te_get_data_length;
	tardata->src.get_data_block = &te_get_data_block;
	tardata->ae = NULL;

	return r;
}

const char *tardata_next(struct tar_data_src *tardata)
{
	int64_t partsize;
	int r;

	r = archive_read_next_header(tardata->ar, &tardata->ae);
	if (r == ARCHIVE_EOF)
		return NULL;

	if (r != ARCHIVE_OK)
	{
		fprintf(stderr, "line %d: archive_read_next_header returned %d\n", __LINE__, r);
		return NULL;
	}

	partsize = archive_entry_size(tardata->ae);
	if (0x100000000LL <= partsize)
	{
		/* FIXME: fix source (and maybe firmware) to deal with files bigger than 4G */
		fprintf(stderr, "line %d: file bigger than 4G won't work %llx\n", __LINE__, (unsigned long long) partsize);
		return NULL;
	}

	return archive_entry_pathname(tardata->ae);
}

int tardata_close(struct tar_data_src *tardata)
{
	archive_read_close(tardata->ar);
	archive_read_finish(tardata->ar);
	return 0;
}

int get_entry_size_in_tar(const char *tarfile, unsigned char *pit, size_t pit_length, size_t *total)
{
	struct tar_data_src tardata;
	const char *filename;
	partition_info *part;
	int r;

	r = tardata_open(&tardata, tarfile);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to open %s\n", __LINE__, tarfile);
		return r;
	}

	*total = 0;

	while (1)
	{
		size_t size = 0;

		filename = tardata_next(&tardata);
		if (!filename)
			break;

		part = find_partition_by_filename(pit, pit_length, filename);
		if (!part)
		{
			fprintf(stderr, "line %d: file %s not found in PIT\n", __LINE__, filename);
			break;
		}

		size = tardata.src.get_data_length(&tardata.src);
		printf("[\x1b[0;32;1m%s\x1b[0m]\t%zuk\n", filename, size/KB);

		/*
		 * Count out if dev_type is FILE because PIT does not have valid info.
		 * Bootloader will do that.
		 */
		if (part->nDevType != PARTITION_DEV_TYPE_FILE)
		{
			unsigned long long part_size = part->dwBlkSize * part->dwBlkLen * KB;

			if (part_size && size > part_size)
			{
				printf("[\x1b[0;31;1mcontent %s too big for partition!\x1b[0m]\n", filename);
				printf("tar %zd bytes block %d*%d*%d bytes\n", size,
					part->dwBlkSize, part->dwBlkLen, KB);
				exit(1);
			}
		}

		*total += size;
	}

	tardata_close(&tardata);

	return 0;
}

int load_single_tarfile(int fd, const char *tarfile, unsigned char *pit, size_t pit_length)
{
	struct tar_data_src tardata;
	const char *filename;
	partition_info *part;
	int r;

	r = tardata_open(&tardata, tarfile);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to open %s\n", __LINE__, tarfile);
		return r;
	}

	while (1)
	{
		filename = tardata_next(&tardata);
		if (!filename)
			break;

		part = find_partition_by_filename(pit, pit_length, filename);
		if (!part)
		{
			fprintf(stderr, "line %d: file %s not found in PIT\n", __LINE__, filename);
			break;
		}

		printf("[\x1b[0;32;1m%s\x1b[0m]\n", filename);
		r = flash_data(fd, &tardata.src, part);
		if (r < 0)
		{
			fprintf(stderr, "line %d: failed to download %s\n", __LINE__, filename);
			break;
		}
	}

	fprintf(stderr, "%s completed\n", tarfile);

	tardata_close(&tardata);

	return r;
}

int transaction_handshake(int fd, size_t *total)
{
	int r;
	unsigned short prot_ver = 0;
	unsigned char plat_ver = 0;
	unsigned char revi_num = 0;
	size_t packet_size = 0;

	/* get target info. */
	r = send_request(fd, RQT_INIT, RQT_INIT_DATA_TARGET, NULL, 0);
	if (r <= 0)
	{
		fprintf(stderr, "RQT_INIT_DATA_TARGET not supported by the target (r=%d)\n", r);
		goto send_total_size;
	}

	prot_ver = (unsigned short)(r >> 16);
	plat_ver = (unsigned char)((r >> 8) & 0xff);
	revi_num = (unsigned char)(r & 0xff);

	fprintf(stderr, "Target reported : \n\tProtocol ver\t%d\n\tPlat. ver\t%d\n\tRev. num\t%d\n",
			prot_ver, plat_ver, revi_num);

	if (prot_ver < 2)
		/* trans unit size = default 128KB */
		goto send_total_size;

	if (prot_ver == 2)
	{
		trans_unit_size = 1*MB;
		goto send_total_size;
	}

	/*
	 * Ver >= 3,
	 * query trans unit size of the target
	 */
	r = send_request(fd, RQT_INIT, RQT_INIT_DATA_XMITSIZE, NULL, 0);
	if (r == -1)
	{
		fprintf(stderr, "RQT_INIT_DATA_XMITSIZE not supported by the target\n");
		goto send_total_size;
	}

	packet_size = (size_t)r;

	/* FIXME : damm.. why report back again? */
	fprintf(stderr, "set transaction unit size to the target ... ");

	r = send_request(fd, RQT_INIT,
			RQT_INIT_DATA_PACKETSIZE,
			(int*)&packet_size, 1);
	if (r)
	{
		fprintf(stderr, "failed (r=%08x)\n", r);
		fprintf(stderr, "use default size.. %zd KB\n", trans_unit_size/KB);

	}
	else
	{
		trans_unit_size = packet_size;
		fprintf(stderr, "success %zd KB\n", trans_unit_size/KB);
	}

send_total_size:

	/* for updating progress bar in the target */
	r = send_request(fd, RQT_INIT, RQT_INIT_DATA_TOTALSIZE, (int*)total, 1);
	if (r)
	{
		fprintf(stderr, "RQT_INIT_DATA_TOTALSIZE, status = %08x\n", r);
		return -1;
	}

	return 0;
}

int process_tar_file_list(const char *portname, char **tarfilelist)
{
	int r;
	char **tfl;
	unsigned char *pit = NULL;
	int pit_length = 0;
	int fd;
	size_t total;

	/* now connect to the target and download the PIT */
	fd = wait_and_open_port(portname);
	if (fd < 0)
	{
		fprintf(stderr, "line %d: failed to open port %s\n", __LINE__, portname);
		return -1;
	}

	r = get_pit_data(fd, &pit, &pit_length);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to fetch pit data %d\n", __LINE__, r);
		return -1;
	}

	tfl = tarfilelist;
	total = 0;
	while (*tfl)
	{
		size_t len = 0;
		printf("\x1b[0;33;1m%s :\x1b[0m\n", *tfl);
		if (get_entry_size_in_tar(*tfl, pit, pit_length, &len) < 0)
		{
			perror("Error");
			return -1;
		}
		total += len;
		tfl++;
	}
	printf("-------------------------\n");
	printf("\t\x1b[0;33;1mtotal\x1b[0m :\t%.2fMB\n\n", (double)total/MB);

	r = transaction_handshake(fd, &total);
	if (r)
		return r;

	while (*tarfilelist)
	{
		fprintf(stderr, "\nLoading files from %s\n\n", *tarfilelist);
		r = load_single_tarfile(fd, *tarfilelist, pit, pit_length);
		if (r < 0)
			fprintf(stderr, "failed to load %s\n", *tarfilelist);
		tarfilelist++;
	}

	r = send_request(fd, RQT_CLOSE, RQT_CLOSE_DATA_END, NULL, 0);
	if (r)
	{
		fprintf(stderr, "\x1b[0;33;1mmissing RQT_CLOSE_DATA_END response from broken bootloader\x1b[0m\n");
	}

	fprintf(stderr, "\nrequest target reboot : ");

	r = request_reboot(fd);
	if (r)
		fprintf(stderr, "\x1b[0;31;1mfailed\x1b[0m\n");
	else
		fprintf(stderr, "\x1b[0;32;1msuccess\x1b[0m\n");

	close(fd);

	return 0;
}

int test_tar_entry(struct data_src *tardata)
{
	static unsigned char *chunk;
	size_t n = 0, size;
	int chunksize;

	chunk = (unsigned char*)malloc(trans_unit_size);
	if (chunk == NULL)
	{
		fprintf(stderr, "Unable to allocate memory.\n");
		return -1;
	}

	size = tardata->get_data_length(tardata);
	while (n < size)
	{
		if (size - n >= trans_unit_size)
			chunksize = trans_unit_size;
		else
			chunksize = size - n;

		chunksize = tardata->get_data_block(tardata, chunk, chunksize);
		if (chunksize <= 0)
		{
			free(chunk);
			return -1;
		}

		if ((!chunksize > 0 && chunksize <= trans_unit_size))
		{
			free(chunk);
			return -1;
		}

		if ((n/chunksize)%4 == 0)
			fprintf(stderr, ".");

		n += chunksize;
	}

	free(chunk);

	return 0;
}

int test_tar_file(const char *tarfile)
{
	struct tar_data_src tardata;
	const char *filename;
	int r = 0;

	fprintf(stderr, "tar %s\n", tarfile);

	r = tardata_open(&tardata, tarfile);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to open %s\n", __LINE__, tarfile);
		return r;
	}

	while (1)
	{
		filename = tardata_next(&tardata);
		if (!filename)
			break;

		fprintf(stderr, " entry %s ", filename);

		r = test_tar_entry(&tardata.src);
		if (r)
		{
			fprintf(stderr, "bad\n");
			break;
		}

		fprintf(stderr, "\n");
	}

	tardata_close(&tardata);

	return r;
}


int test_tar_file_list(char **tarfilelist)
{
	int r;

	while (*tarfilelist)
	{
		char *tarfile = *tarfilelist;

		r = test_tar_file(tarfile);
		if (r < 0)
			fprintf(stderr, "failed to load %s\n", *tarfilelist);
		tarfilelist++;
	}

	return 0;
}

char *ipl_name[] = {
	"ipl+recovery",
	"s-boot-onenand",
	"s-boot-mmc",
	"MLO",
	NULL,
};

char *sbl_name[] = {
	"u-boot",
	"u-boot-onenand",
	"u-boot-mmc",
	NULL,
};

partition_info* find_ipl(unsigned char *pit, int pit_length, char **pit_list, const char *bootloader, size_t *sz)
{
	int r;
	int i;
	int found;
	partition_info* ipl_part;
	const char *entry;
	struct tar_data_src tardata;

	found = 0;
	ipl_part = NULL;
	entry = NULL;
	for (i = 0; pit_list[i] != NULL; i++)
	{
		ipl_part = find_partition_by_name(pit, pit_length, pit_list[i]);
		if (!ipl_part)
			continue;

		r = tardata_open(&tardata, bootloader);
		if (r < 0)
		{
			fprintf(stderr, "line %d: failed to open %s (%d)\n", __LINE__, bootloader, r);
			return NULL;
		}

		for (;;)
		{
			entry = tardata_next(&tardata);
			if (!entry)
				break;

			if (!strcmp(entry, ipl_part->szFileName))
			{
				*sz += tardata.src.get_data_length(&tardata.src);
				found = 1;
				break;
			}
		}

		tardata_close(&tardata);

		if (found)
			break;
	}

	if (!found)
	{
		if (!ipl_part)
			fprintf(stderr, "line %d: no ipl partition in PIT file?\n", __LINE__);
		if (!entry)
			fprintf(stderr, "line %d: %s not in tar file?\n", __LINE__, ipl_part->szFileName);

		ipl_part = NULL;
	} else
		printf("%s found\n", ipl_part->szName);

	return ipl_part;
}

/*
 * Should take care to download uboot when the partition table is changed
 * Not doing this might leave the user with a phone that can't be flashed
 * Requiring somebody with a JTAG setup to fix it...
 *
 * Care here means:
 * - check that the tar file we're going to download actually has uboot in it
 */
int process_pit_update(const char *portname, const char *pitfile, const char *bootloader)
{
	int pit, fd, r, pitsize;
	void *pitaddr;
	partition_info* ipl_part, *sbl_part;
	struct tar_data_src tardata;
	const char *entry;
	size_t total = 0;

	fprintf(stderr, "loading PIT file and bootloader\n");

	/* get the size of the pitfile */
	pit = open(pitfile, O_RDONLY);
	if (pit < 0)
	{
		fprintf(stderr, "line %d: failed to open pitfile %s\n", __LINE__, pitfile);
		return 1;
	}

	pitsize = lseek(pit, 0, SEEK_END);
	if (pitsize == -1)
	{
		fprintf(stderr, "line %d: failed to seek pit\n", __LINE__);
		return -1;
	}

	lseek(pit, 0, SEEK_SET);

	if (opt_verbose)
		fprintf(stderr, "pit file size is %d bytes\n", pitsize);

	/* map and send the pitfile */
	pitaddr = mmap(NULL, pitsize, PROT_READ, MAP_PRIVATE, pit, 0);
	if (pitaddr == (void*) -1)
	{
		fprintf(stderr, "line %d: failed to map file\n", __LINE__);
		return -1;
	}

	ipl_part = find_ipl(pitaddr, pitsize, ipl_name, bootloader, &total);
	if (!ipl_part)
	{
		fprintf(stderr, "line %d: no ipl partition in PIT file?\n", __LINE__);
		return -1;
	}

	sbl_part = find_ipl(pitaddr, pitsize, sbl_name, bootloader, &total);
	if (!sbl_part)
	{
		fprintf(stderr, "line %d: no 2nd boot loader partition in PIT file?\n", __LINE__);
		return -1;
	}

	/* ok, all is good with the data we're given, this is point of no return */

	fd = wait_and_open_port(portname);
	if (fd < 0)
	{
		fprintf(stderr, "line %d: failed to open port %s\n", __LINE__, portname);
		return -1;
	}

	r = transaction_handshake(fd, &total);
	if(r)
		return r;

	r = write_pit_data(fd, pitaddr, pitsize);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed write pit file\n", __LINE__);
		return -1;
	}

	r = tardata_open(&tardata, bootloader);
	if (r < 0)
	{
		fprintf(stderr, "line %d: failed to open %s\n", __LINE__, bootloader);
		return r;
	}

	while (1)
	{
		entry = tardata_next(&tardata);
		if (!entry)
			break;

		if (!strcmp(entry, ipl_part->szFileName))
		{
			r = flash_data(fd, &tardata.src, ipl_part);
			if (r < 0)
			{
				fprintf(stderr, "line %d: failed to download %s - maybe bad news :(\n", __LINE__, ipl_part->szFileName);
				return -1;
			}
		}

		if (!strcmp(entry, sbl_part->szFileName))
		{
			r = flash_data(fd, &tardata.src, sbl_part);
			if (r < 0)
			{
				fprintf(stderr, "line %d: failed to download %s - maybe bad news :(\n", __LINE__, sbl_part->szFileName);
				return -1;
			}
		}

	}

	request_reboot(fd);

	tardata_close(&tardata);

	munmap(pitaddr, pitsize);
	close(pit);

	return 0;
}

void usage(const char *exename)
{
	fprintf(stderr, "%s: [-t] [-v] [-d port] [-p pitfile] [-getpit pitfile] [tarfile] [tarfile] ...\n", exename);
	exit(1);
}

int main(int argc, char **argv)
{
	const char *exename, *portname, *pitfile, *getpitfile, *bootloader;
	int opt;

	exename = argv[0];

	opt = 1;

	pitfile = NULL;
	bootloader = NULL;
	getpitfile = NULL;

	portname = NULL;

	printf("\n");
	printf("Linux Odin3 downloader, version %s (%s%s) built %s\n",
		PACKAGE_VERSION, git_version, git_dirty ? "-dirty" : "", build_date);
	printf("Authors: mj.mccormack@samsung.com, Minho Ban <mhban@samsung.com>\n\n");

	while (opt < argc)
	{
		/* check if we're verbose */
		if (!strcmp(argv[opt], "-v"))
		{
			opt_verbose = 1;
			opt++;
			continue;
		}

		if (!strcmp(argv[opt], "-t"))
		{
			opt_test = 1;
			opt++;
			continue;
		}

		if (!strcmp(argv[opt], "-p"))
		{
			if ((opt+3) != argc)
			{
				fprintf(stderr, "Load PIT with: lodin -p file.pit bootloader.tar\n");
				fprintf(stderr, "The bootloader may be in the kernel image.\n");
				exit(1);
			}

			pitfile = argv[opt+1];
			bootloader = argv[opt+2];
			opt += 3;
			continue;
		}

		if (!strcmp(argv[opt], "-d") && (opt+1) < argc)
		{
			portname = argv[opt+1];
			opt += 2;
			continue;
		}

		if (!strcmp(argv[opt], "-getpit") && (opt+1) < argc)
		{
			getpitfile = argv[opt+1];
			opt += 2;
			continue;
		}

		break;
	}

	if (opt_test)
		return test_tar_file_list(&argv[opt]);

	if (getpitfile)
		return get_pit_file(portname, getpitfile);

	if (pitfile)
		return process_pit_update(portname, pitfile, bootloader);

	if (opt < argc)
		return process_tar_file_list(portname, &argv[opt]);

	usage(exename);
	return 0;
}

